#include "qffmpeg.h"

// Открывает видеофайл, инициализирует кодеки, получает информацию о потоках
TFfmpeg::TFfmpeg(QString videoFile){
	error=0;
	av_log_set_level(AV_LOG_QUIET);
	avcodec_register_all();
	av_register_all();

	pAVFormatContext = avformat_alloc_context();
	if (avformat_open_input (&pAVFormatContext, videoFile.toLocal8Bit(), NULL, NULL)!=0){
		QTextStream(stdout)<<"Error open file:"<<videoFile<<"\n";
		error=1;
		return;
		}
	else if (avformat_find_stream_info (pAVFormatContext, NULL) < 0){
		QTextStream(stdout)<<"Not found streams in"<<videoFile<<"\n";
		error=1;
		return;
		}

// Продолжительность
	duration = QTime(0,0,0,0).addMSecs (pAVFormatContext->duration / 1000).toString("hh:mm:ss");
// Общий битрейт всех потоков
	fullBitrate = QString::number (pAVFormatContext->bit_rate / 1000);
// Количество потоков
	streamsCount = pAVFormatContext->nb_streams;
// Строка основных параметров
	generalInfo = duration+", "+QObject::tr("Streams:")+" "+QString::number (streamsCount)+", "+QObject::tr("Full bitrate:")+" "+fullBitrate+" "+QObject::tr("kb/s");

// Проверка наличия видеопотока и получение его номера
	vStreamNumber=-1;
	for (uint i=0; i < pAVFormatContext->nb_streams;i++)
		if (pAVFormatContext->streams[i]->codec->codec_type == AVMEDIA_TYPE_VIDEO){
			vStreamNumber=i;
			break;
			}
	if (vStreamNumber == -1){
		QTextStream(stdout)<<"Not video stream in"<<videoFile<<"\n";
		error=1;
		return;
		}

// Проверка наличия в ffmpeg кодека для видеопотока
	AVStream *vStream = pAVFormatContext->streams[vStreamNumber];
	vCodecContext = vStream->codec;
	vCodec = avcodec_find_decoder (vCodecContext->codec_id);
	if (avcodec_open2 (vCodecContext, vCodec, NULL)){
		QTextStream(stdout)<<"Not open video codec in"<<videoFile<<"\n";
		error=1;
		return;
		}

// Частота кадров
	vFps = QString::number ((double) vStream->r_frame_rate.num / vStream->r_frame_rate.den);
// Кодек
	vCodecName = QString(vCodec->name).toUpper();
// Размер исходного кадра
	vWidth = vCodecContext->width;
	vHeight = vCodecContext->height;
// Размер исходного кадра в текстовом виде
	vResolution = QString::number (vWidth) + "x" + QString::number (vHeight);
// Продолжительность
	iDuration = pAVFormatContext->duration;
	
// Вычисление Display Aspect Ratio (DAR). Соотношение сторон выводимого на экран кадра
	dar = double (vWidth) / double (vHeight);
	if (vStream->sample_aspect_ratio.num && av_cmp_q(vStream->sample_aspect_ratio, vStream->codec->sample_aspect_ratio)){
		AVRational display_aspect_ratio;
		av_reduce (&display_aspect_ratio.num, &display_aspect_ratio.den, vWidth * vStream->sample_aspect_ratio.num, vHeight * vStream->sample_aspect_ratio.den, 1024*1024);
		dar = double (display_aspect_ratio.num) / double (display_aspect_ratio.den);
		} else {
			if (vCodecContext->sample_aspect_ratio.num) {
			AVRational display_aspect_ratio;
			av_reduce (&display_aspect_ratio.num, &display_aspect_ratio.den, vWidth * vCodecContext->sample_aspect_ratio.num, vHeight * vCodecContext->sample_aspect_ratio.den, 1024 * 1024);
			dar = double (display_aspect_ratio.num) / double (display_aspect_ratio.den);
			}
		}
		
// Проверка наличия аудиопотока и получение его номера
	aStreamNumber=-1;
	for (uint i=0; i < pAVFormatContext->nb_streams;i++)
		if (pAVFormatContext->streams[i]->codec->codec_type == AVMEDIA_TYPE_AUDIO){
			aStreamNumber=i;
			break;
			}
	if (aStreamNumber == -1) QTextStream(stdout)<<"Not audio stream in\n"<<videoFile<<endl;
	else{
		aCodecContext = pAVFormatContext->streams[aStreamNumber]->codec;
		aCodec = avcodec_find_decoder (aCodecContext->codec_id);
		if (avcodec_open2 (aCodecContext, aCodec, NULL)) QTextStream(stdout)<<"Not open audio codec in"<<videoFile<<"\n"<<endl;
		}

	if (vCodecContext->codec_tag > 0){
		char vCodecTagC[32];
		av_get_codec_tag_string (vCodecTagC, sizeof(vCodecTagC), vCodecContext->codec_tag);
		vCodecTag = QString (vCodecTagC);
		}
	char aChC[32];
	if (aStreamNumber != -1){
		av_get_channel_layout_string (aChC, sizeof(aChC), aCodecContext->channels, aCodecContext->channel_layout);
// Количество каналов
		aChannels = QString (aChC);
// Частота дискретизации
		aSamples = QString::number ((double)aCodecContext->sample_rate / 1000);
// Битрейт
		aBitrate = QString::number (get_bit_rate (aCodecContext)/1000);
// Кодек
		aCodecName = QString(aCodec->name).toUpper();
		}
	}

// Вычисляет битрейт потока (пока только для звука, и то, не всегда)
int TFfmpeg::get_bit_rate(AVCodecContext *ctx){
	int bit_rate = 0;
	int bits_per_sample;

	switch(ctx->codec_type) {
		case AVMEDIA_TYPE_VIDEO:
		case AVMEDIA_TYPE_DATA:
		case AVMEDIA_TYPE_SUBTITLE:
		case AVMEDIA_TYPE_ATTACHMENT:
			bit_rate = ctx->bit_rate;
			break;
		case AVMEDIA_TYPE_AUDIO:
			bits_per_sample = av_get_bits_per_sample(ctx->codec_id);
			if (bits_per_sample) bit_rate = ctx->sample_rate * ctx->channels * bits_per_sample;
				else bit_rate = ctx->bit_rate;
			break;
		default:
			break;
		}
	return bit_rate;
}

// Получение строки с информацией о видео потоке
QString TFfmpeg::getVInfo(){
	QString vct = ", ";
	if ((vCodecContext->codec_tag > 0) &&
		(vCodecTag != vCodecName) &&
		(vCodecTag.mid(0,1) != "[") &&
		(vCodecTag.mid (vCodecTag.length()-1,1) != "]"))
			vct = " ("+vCodecTag+"), ";
	return vCodecName + vct + vResolution + ", " + vFps + " " + QObject::tr("fps");
}

// Получение строки с информацией об аудио потоке
QString TFfmpeg::getAInfo(){
	QString ret="";
	if (aStreamNumber != -1){
		QString br = "";
		if (aBitrate != "0") br = ", " + aBitrate + " " + QObject::tr("kb/s");
		ret = aCodecName + ", " + aChannels.toUpper() + ", " + aSamples + " " + QObject::tr("kHz") + br;
		}
	else ret = QObject::tr("Not present");
	return ret;
	}

// Закрытие видеофайла
TFfmpeg::~TFfmpeg(){
	avformat_close_input (&pAVFormatContext);
}

// Получение картинки в QImage кадра из видеофайла в позиции ms миллисекунд
QImage TFfmpeg::getImage(uint ms){
	QImage img = QImage(vCodecContext->width, vCodecContext->height, QImage::Format_RGB888);
	AVFrame *pFrame, *pFrameRGB;
	AVPacket packet;
	int numBytes;
	uint8_t *buffer;
	int frameFinished;

	pFrame = avcodec_alloc_frame();
	pFrameRGB = avcodec_alloc_frame();

	numBytes = avpicture_get_size (PIX_FMT_RGB24, vCodecContext->width, vCodecContext->height);
	buffer = (uint8_t *) av_malloc (numBytes * sizeof(uint8_t));
	avpicture_fill ((AVPicture *) pFrameRGB, buffer, PIX_FMT_RGB24,	vCodecContext->width, vCodecContext->height);

	int frameNum = av_rescale (ms,pAVFormatContext->streams[vStreamNumber]->time_base.den, pAVFormatContext->streams[vStreamNumber]->time_base.num) / 1000;
	avformat_seek_file (pAVFormatContext, vStreamNumber, 0, frameNum, frameNum, AVSEEK_FLAG_FRAME);
	avcodec_flush_buffers (vCodecContext);

	while (av_read_frame (pAVFormatContext, &packet) >= 0){
		if (packet.stream_index == vStreamNumber){
			avcodec_decode_video2 (vCodecContext, pFrame, &frameFinished, &packet);
			if (frameFinished){
				struct SwsContext *context;
				context = sws_getCachedContext (0, vCodecContext->width, vCodecContext->height, vCodecContext->pix_fmt, vCodecContext->width,vCodecContext->height, PIX_FMT_RGB24, SWS_BICUBIC, 0, 0, 0);
				sws_scale (context, (uint8_t**) pFrame->data, pFrame->linesize, 0, vCodecContext->height, pFrameRGB->data, pFrameRGB->linesize);
				for (int j=0; j<vCodecContext->height; j++) memcpy (img.scanLine (j), pFrameRGB->data[0] + j * pFrameRGB->linesize[0], vCodecContext->width * 3);
				break;
				}
			}
//		av_free_packet (&packet);
		}
	return img;
}
